#define pam_sm_acct_mgmt _pam_unix_sm_acct_mgmt
#define pam_sm_authenticate _pam_unix_sm_authenticate
#define pam_sm_setcred _pam_unix_sm_setcred
#define pam_sm_chauthtok _pam_unix_sm_chauthtok
#define pam_sm_open_session _pam_unix_sm_open_session
#define pam_sm_close_session _pam_unix_sm_close_session
